"""Dialog box widgets for various GUI functions"""

from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5.QtCore import Qt
import sys
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import re

from pandasgui.utility import flatten_multiindex


class DialogGeneric(QtWidgets.QDialog):
    """
    This widget allows the user to pick columns from DataFrames and pass the choices to a wrapper class with getters

    Args:
        dataframes:
        destination_names:
        destination_options:
        parent:
        default:
    """

    def __init__(self, dataframes, destination_names, title, parent=None, default_df=None):

        super().__init__(parent)

        self.dataframes = dataframes
        self.setWindowTitle(title)

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        # Create column picker
        self.columnPicker = Picker(destination_names=destination_names)

        # Create DataFrame picker dropdown
        self.dataframePicker = QtWidgets.QComboBox()
        # Fill the dataframe picker with the names of dataframes
        for df_name in dataframes.keys():
            self.dataframePicker.addItem(df_name)

        # Set default selection and trigger currentIndexChanged
        index = self.dataframePicker.findText(default_df)
        if index != -1:
            self.dataframePicker.setCurrentIndex(index)
        else:
            self.dataframePicker.setCurrentIndex(0)
        self.dataframeChanged()

        # Connect signals
        self.dataframePicker.currentIndexChanged.connect(self.dataframeChanged)

        # Add button
        btnFinish = QtWidgets.QPushButton("Finish")
        btnFinish.clicked.connect(self.finish)
        btnReset = QtWidgets.QPushButton("Reset")
        btnReset.clicked.connect(self.initColumnPicker)
        buttonLayout = QtWidgets.QHBoxLayout()
        spacer = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        buttonLayout.addSpacerItem(spacer)
        buttonLayout.addWidget(btnReset)
        buttonLayout.addWidget(btnFinish)

        # Add all to layout
        layout.addWidget(self.dataframePicker)
        layout.addWidget(self.columnPicker)
        layout.addLayout(buttonLayout)

        self.resize(640, 480)

    def finish(self):
        print('Finish (This should be overridden)')

    def getDestinationItems(self):
        return self.columnPicker.getDestinationItems()

    def getDataFrameName(self):
        return self.dataframePicker.itemText(self.dataframePicker.currentIndex())

    def getDataFrame(self):
        df_name = self.dataframePicker.itemText(self.dataframePicker.currentIndex())
        return self.dataframes[df_name]['dataframe']

    def dataframeChanged(self):
        print('dfchanged')
        # Get the name of the selected dataframe from the dataframePicker
        selected_df_name = self.dataframePicker.itemText(self.dataframePicker.currentIndex())
        self.selected_df = self.dataframes[selected_df_name]['dataframe'].copy()

        self.initColumnPicker()

    def initColumnPicker(self):
        self.columnPicker.clearDestinationItems()
        column_names = list(self.selected_df.columns)
        self.columnPicker.setSourceItems(column_names)


# Widget for selecting DataFrame columns from the SourceList into multiple destination lists (DestList) for usage in
# the dialog function. For example the destinations could be XVariable, Y-Variables, ColorBy for the ScatterPlot dialog
class Picker(QtWidgets.QWidget):
    def __init__(self, destination_names=['Default Dest'], column_names=['Default Cols']):
        super().__init__()

        # Set up widgets and layout
        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        self.columnSource = RegexSourceList(column_names)
        self.destinations = []
        for name in destination_names:
            self.destinations.append(DestList(name))

        # Add items to layout
        self.destLayout = QtWidgets.QVBoxLayout()
        for dest in self.destinations:
            self.destLayout.addWidget(dest)
        layout.addWidget(self.columnSource)
        layout.addLayout(self.destLayout)

    def setSourceItems(self, items=None):
        self.columnSource.setItems(items)

    def resetItems(self):
        # Clear list
        self.columnSource.setItems()

    def clearDestinationItems(self):
        # Clear tree
        for dest in self.destinations:
            dest.clear()

    # Return a dict of the items in the destination trees
    def getDestinationItems(self):
        items = {}

        for dest in self.destinations:
            items[dest.title] = dest.getItems()
        return items


# Though the content is a flat list this is implemented as a QTreeWidget for some additional functionality like column
# titles and multiple columns
class DestList(QtWidgets.QTreeWidget):
    def __init__(self, title='Variable', parent=None):
        super().__init__(parent)
        self.title = title
        self.setHeaderLabels([title])

        # Tree settings
        self.setDragDropMode(self.DragDrop)
        self.setSelectionMode(self.ExtendedSelection)
        self.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.setAcceptDrops(True)

        # Remove items by double clicking them
        self.doubleClicked.connect(self.removeSelectedItems)

    def removeSelectedItems(self):
        for item in self.selectedItems():
            self.invisibleRootItem().removeChild(item)

    def getItems(self):
        """
        Returns:
            [str]: a list of the items in the tree
        """
        items = []
        for i in range(self.topLevelItemCount()):
            treeItem = self.topLevelItem(i)
            items.append(treeItem.text(0))
        return items


class RegexSourceList(QtWidgets.QWidget):
    def __init__(self, item_list, parent=None):
        super().__init__(parent=parent)
        self.item_list = item_list

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        self.searchBox = QtWidgets.QLineEdit()
        self.list = SourceList(parent=parent)
        self.setItems(item_list)

        self.searchBox.textChanged.connect(self.filter)

        layout.addWidget(self.searchBox)
        layout.addWidget(self.list)

    def filter(self):
        filteredItems = [item for item in self.item_list if re.search(self.searchBox.text().lower(), item.lower())]
        self.list.setItems(filteredItems)

    def setItems(self, items):
        self.item_list = items
        self.list.setItems(items)


class SourceList(QtWidgets.QListWidget):
    '''
    A QListWidget that shows the list of column names for a dataframe given as columnNames
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        # Settings
        self.setDragDropMode(self.DragDrop)
        self.setSelectionMode(self.ExtendedSelection)
        self.setDefaultDropAction(QtCore.Qt.CopyAction)
        self.setAcceptDrops(True)

    # Allow dropping to this list but keep it unchanging by resetting it every time this happens. So the item will just
    # be removed from the DestList it was dragged from.
    def dropEvent(self, event):
        itemsTextList = self.getItems()

        # Default action
        QtWidgets.QListWidget.dropEvent(self, event)

        self.setItems(itemsTextList)

    def getItems(self):
        return [str(self.item(i).text()) for i in range(self.count())]

    # This gets called when an item gets dragged from DestList to SourceList or vice versa
    def setItems(self, items):
        self.clear()
        for name in items:
            self.addItem(name)


# %% Specific dialogs that inherit GenericDialog


###
class PivotDialog(DialogGeneric):
    def __init__(self, dataframes, default_df=None, parent=None):
        super().__init__(dataframes, destination_names=['index', 'columns', 'values'], default_df=default_df,
                         title='Pivot', parent=parent)

        self.show()

    def finish(self):
        dict = self.getDestinationItems()
        df = self.getDataFrame()
        df_name = self.getDataFrameName()

        try:
            index = dict['index']
            columns = dict['columns']
            values = dict['values']

            from pandasgui import show
            pivot_table = df.pivot_table(values, index, columns)

            self.gui.add_dataframe(df_name + "_pivot", pivot_table, parent_name=df_name)

        except Exception as e:
            print(e)


##

class Categorizer(QtWidgets.QDialog):
    def __init__(self, df, column):
        super().__init__()

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.df = df
        from pandasgui import show
        self.column_name = column
        self.names = QtWidgets.QLineEdit()
        self.names.textChanged.connect(self.makePicker)

        self.picker = Picker(['col1', 'col2', 'col3'], df[column].astype(str).unique())

        # Add button
        btnFinish = QtWidgets.QPushButton("Finish")
        btnFinish.clicked.connect(self.finish)
        btnReset = QtWidgets.QPushButton("Reset")
        btnReset.clicked.connect(self.finish)
        buttonLayout = QtWidgets.QHBoxLayout()
        spacer = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        buttonLayout.addSpacerItem(spacer)
        buttonLayout.addWidget(btnReset)
        buttonLayout.addWidget(btnFinish)

        # Add all to layout
        self.layout.addLayout(buttonLayout)
        self.layout.addWidget(self.names)
        self.layout.addWidget(self.picker)
        self.show()

    def makePicker(self):
        self.layout.removeWidget(self.picker)
        self.df[self.column_name].astype(str).unique()
        self.picker = Picker(self.names.text().split(','), self.df[self.column_name].astype(str).unique())
        print(self.df[self.column_name].astype(str).unique())
        self.layout.addWidget(self.picker)

    def finish(self):
        dict = self.picker.getDestinationItems()

        try:
            print(self.picker.getDestinationItems())
            mapping = {
                'Columbia': 'South-America',
                'Ecuador': 'South-America',
                'Peru': 'South-America',
                'South-Africa': 'Africa',
                'Namibia': 'Africa',
            }

            results = self.picker.getDestinationItems()
            mapping = {}
            for key in results.keys():
                for value in results[key]:
                    mapping[value] = key

            print(mapping)
            self.df[self.column_name + ' Categorized'] = self.df[self.column_name].astype(str).replace(mapping)
            from pandasgui import show
            show(self.df)

        except Exception as e:
            print(e)


class ScatterDialog(DialogGeneric):
    def __init__(self, dataframes, default_df=None, parent=None):
        super().__init__(dataframes, destination_names=['X Variable', 'Y Variable', 'Color By'], default_df=default_df,
                         title='Scatter Plot', parent=parent)

        self.show()

    def finish(self):
        dict = self.getDestinationItems()
        df = self.getDataFrame()

        try:
            x = dict['X Variable'][0]
            y = dict['Y Variable'][0]
            c = dict['Color By'][0]
        except IndexError:
            c = None

        sns.scatterplot(x, y, c, data=df)
        plt.show()


if __name__ == '__main__':
    pass



"""Widgets for viewing figures"""

from PyQt5 import QtWidgets
import matplotlib
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import random

matplotlib.use('Qt5Agg')

class TabbedFigureViewer(QtWidgets.QWidget):
    """
    Displays a list of matplotlib Figures in a tabbed viewer
    Args:
        figs (List): List of matplotlib Figures
    """

    def __init__(self, figs=[]):

        super().__init__()

        self.setWindowTitle("Tabbed Image Viewer")
        self.figs = figs
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        # Create tab widget
        self.tab_widget = QtWidgets.QTabWidget()
        self.tab_widget.setMovable(True)
        self.tab_widget.setTabsClosable(True)

        # Add figures to tab widget
        for fig in figs:
            self.add_tab(fig)

        # Just some button connected to `plot` method
        self.button = QtWidgets.QPushButton('To PowerPoint')
        self.button.clicked.connect(self.to_powerpoint)

        # Add widgets to layout
        layout.addWidget(self.tab_widget)
        layout.addWidget(self.button)

        self.show()

        # Center window on screen
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move(int((screen.width() - size.width()) / 2), int((screen.height() - size.height()) / 2))

    def add_tab(self, fig):
        """Add a new tab containing a FigureViewer showing fig"""
        image_viewer = FigureViewer(fig)
        # Adds them to the tab_view
        tab = QtWidgets.QWidget()
        tab_layout = QtWidgets.QVBoxLayout()
        tab.setLayout(tab_layout)
        tab_layout.addWidget(image_viewer)
        try:
            fig_title = fig.texts[0].get_text()
        except IndexError:
            fig_title = f"untitled"
        self.tab_widget.addTab(tab, fig_title)

    def to_powerpoint(self):
        # Write images to ppt
        import os
        filename = 'test.ppt'
        try:
            os.remove(filename)
        except OSError:
            pass
        for fig in self.figs:
            to_ppt_slide(fig, filename, append=True)
        os.system("start " + filename)
        win = TabbedFigureViewer(figs)


class FigureViewer(QtWidgets.QWidget):
    """
    Displays a matplotlib figure and toolbar in a widget

    Args:
        fig (Figure): matplotlib Figure to show
    """

    def __init__(self, fig=Figure()):
        super().__init__()

        self.canvas = FigureCanvasQTAgg(fig)
        self.toolbar = NavigationToolbar2QT(self.canvas, self)

        # Just some button connected to `plot` method
        self.button = QtWidgets.QPushButton('Plot Random')
        self.button.clicked.connect(self.setRandomImage)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.toolbar)
        self.layout.addWidget(self.canvas)
        # self.layout.addWidget(self.button)

        self.toolbar.setStyleSheet('background: transparent;')

    def setFigure(self, fig):
        """Replace the current figure shown in the widget with fig"""
        plt.close(self.canvas.figure)

        new_canvas = FigureCanvasQTAgg(fig)
        self.layout.replaceWidget(self.canvas, new_canvas)

        new_toolbar = NavigationToolbar2QT(new_canvas, self)
        self.layout.replaceWidget(self.toolbar, new_toolbar)

        self.toolbar.setParent(None)
        self.toolbar = new_toolbar
        self.canvas = new_canvas

    def setRandomImage(self):
        """Plot some random data and show it in the widget"""
        data = [random.random() for i in range(10)]
        plt.figure()
        plt.plot(data, '*-')
        self.setFigure(plt.gcf())
        self.canvas.draw()


def to_ppt_slide(fig, file_path, append=False, padding=0.5):
    from io import StringIO, BytesIO
    import pptx
    from pptx import Presentation
    from pptx.util import Inches

    # Create in-memory image stream and save figure to it
    image_stream = BytesIO()
    fig.savefig(image_stream)

    if append:
        try:
            # Try opening the file if it already exists
            prs = Presentation(file_path)
        except pptx.exc.PackageNotFoundError:
            prs = Presentation()
    else:
        prs = Presentation()
    # Create a new slide with the blank template
    blank_slide_layout = prs.slide_layouts[6]
    slide = prs.slides.add_slide(blank_slide_layout)

    # Center image without changing its aspect ratio
    slide_width = prs.slide_width.inches - 2 * padding
    slide_height = prs.slide_height.inches - 2 * padding
    fig_width, fig_height = fig.get_size_inches()
    if (fig_width / slide_width) > (fig_height / slide_height):
        # Image fits slide horizontally and must be scaled down vertically
        width = slide_width
        height = width * fig_height / fig_width
        top = padding + (slide_height - height) / 2
        left = padding
    else:
        # Image fits slide vertically and must be scaled down horizontally
        height = slide_height
        width = height * fig_width / fig_height
        left = padding + (slide_width - width) / 2
        top = padding

    # Convert from EMU to inches
    left = Inches(left)
    top = Inches(top)
    height = Inches(height)
    width = Inches(width)

    pic = slide.shapes.add_picture(image_stream, left, top, height=height, width=width)

    prs.save(file_path)


def ppt_section_slide(title, subtitle, file_path):
    from pptx import Presentation
    try:
        prs = Presentation(file_path)
    except:
        prs = Presentation(r'C:\Users\adrose\Desktop\AMD PowerPoint Template.pptx')
    picture_slide_layout = prs.slide_layouts[0]
    slide = prs.slides.add_slide(picture_slide_layout)

    # for x in slide.placeholders:
    #     print('%d %s' % (x.placeholder_format.idx, x.name))

    title_placeholder = slide.placeholders[0]
    subtitle_placeholder = slide.placeholders[1]

    title_placeholder.text = title
    subtitle_placeholder.text = subtitle

    prs.save(file_path)


if __name__ == '__main__':
    from pandasgui.datasets import iris, flights, multi, pokemon

    app = QtWidgets.QApplication([])

    figs = []
    for df in [iris, flights, multi]:
        df.plot()
        figs.append(plt.gcf())

    win = FigureViewer(figs[0])
    win.show()

    win2 = TabbedFigureViewer(figs)
    win2.show()

    app.exec_()



import os, sys
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QUrl
from plotly.io import to_html
import tempfile

# If QtWebEngineWidgets is imported while a QApplication instance already exists it will fail, so we have to
try:
    from PyQt5 import QtWebEngineWidgets
except ImportError as e:
    if e.msg == "QtWebEngineWidgets must be imported before a QCoreApplication instance is created":
        print("Killing QApplication to reimport QtWebEngineWidgets")
        from PyQt5 import sip

        app = QApplication.instance()
        app.quit()
        sip.delete(app)
        del app
        from PyQt5 import QtWebEngineWidgets

        # Without remaking the QApplication you will get an unrecoverable crash on the next attempted usage of it
        app = QApplication([])
    else:
        raise e


class PlotlyViewer(QtWebEngineWidgets.QWebEngineView):
    def __init__(self, fig=None):
        super().__init__()

        if fig:
            self.set_figure(fig)

        self.resize(700, 600)
        self.setWindowTitle("Plotly Viewer")

    def set_figure(self, fig):
        config = {'responsive': True}
        self.temp_file = tempfile.TemporaryFile(mode='w', suffix='.html')
        self.temp_file.write(to_html(fig, config=config))
        self.temp_file.seek(0)
        self.load(QUrl.fromLocalFile(self.temp_file.name))

if __name__ == "__main__":
    # Create a QApplication instance or use the existing one if it exists
    app = QApplication(sys.argv)

    import numpy as np
    import plotly.graph_objs as go
    from pandasgui.utility import fix_ipython, fix_pyqt
    fix_ipython()
    fix_pyqt()

    fig = go.Figure()
    fig.add_scatter(x=np.random.rand(100), y=np.random.rand(100), mode='markers',
                    marker={'size': 30, 'color': np.random.rand(100), 'opacity': 0.6,
                            'colorscale': 'Viridis'});

    pv = PlotlyViewer(fig)
    # app.exec_()


"""Reusable PyQt widgets"""

__all__ = ['DataFrameExplorer', 'DataFrameViewer', 'FigureViewer', 'TabbedFigureViewer', 'PivotDialog',
           'ScatterDialog', 'PlotlyViewer', 'GraphBuilder', 'FindToolbar']

from pandasgui.widgets.find_toolbar import FindToolbar
from pandasgui.widgets.image_viewer import FigureViewer, TabbedFigureViewer
from pandasgui.widgets.plotly_viewer import PlotlyViewer
from pandasgui.widgets.dialogs import PivotDialog, ScatterDialog
from pandasgui.widgets.graph_builder import GraphBuilder

from pandasgui.widgets.dataframe_viewer import DataFrameViewer
from pandasgui.widgets.dataframe_explorer import DataFrameExplorer



"""DataFrameExplorer"""

from PyQt5 import QtWidgets
import pandas as pd
import sys
import matplotlib.pyplot as plt
import seaborn as sns
from pandasgui.widgets import DataFrameViewer
from pandasgui.widgets import FigureViewer
from pandasgui.widgets import GraphBuilder
import traceback

class DataFrameExplorer(QtWidgets.QTabWidget):
    """
    This is a QTabWidget for analyzing a single DataFrame where the first tab is a DataFrameViewer widget

    Args:
        df (DataFrame): The DataFrame to display
    """

    def __init__(self, df):

        super().__init__()

        df = df.copy()
        self.df = df

        # DataFrame tab
        self.dataframe_tab = DataFrameViewer(self.df)
        self.addTab(self.dataframe_tab, "DataFrame")

        # Statistics tab
        try:
            self.statistics_tab = self.make_statistics_tab(df)
            self.addTab(self.statistics_tab, "Statistics")
        except:
            traceback.print_exc()

        # Histogram tab
        histogram_tab = GraphBuilder(df)
        self.addTab(histogram_tab, "Histogram")

    def make_statistics_tab(self, df):
        stats_df = pd.DataFrame({
            'Type': df.dtypes.replace('object', 'string'),
            'Count': df.count(),
            'Mean': df.mean(numeric_only=True),
            'StdDev': df.std(numeric_only=True),
            'Min': df.min(numeric_only=True),
            'Max': df.max(numeric_only=True),
        })
        w = DataFrameViewer(stats_df)
        w.setAutoFillBackground(True)
        return w



# Examples
if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    from pandasgui.datasets import iris, flights, multi, pokemon

    # Create and show widget
    dfe = DataFrameExplorer(flights)
    dfe.show()

    sys.exit(app.exec_())



# This file isn't used. Keeping it for reference in case I want to re-add the separate process functionality

"""Creates a non-blocking instance of the Pandas GUI in a separate process. Works in both script & interactive mode"""

import multiprocessing


def start_gui(**kwargs):
    import sys
    from PyQt5 import QtWidgets
    from pandasgui.gui import PandasGUI

    # Set up QApplication
    app = QtWidgets.QApplication.instance()
    if not app:
        app = QtWidgets.QApplication(sys.argv)

    # Make GUi
    win = PandasGUI(block=True, **kwargs)

    app.exec_()


def show_nonblocking(**kwargs):
    thread = multiprocessing.Process(target=start_gui, kwargs=kwargs)
    thread.start()


if __name__ == '__main__':
    from pandasgui import show

    from pandasgui.datasets import iris, flights, multi, pokemon

    show(iris, flights, multi, pokemon)



import pandas as pd
import seaborn as sns

def run_forecasts()

def generate_dashboard():
    df = pd.read_csv("raw_yield_data.csv")

    #region Add columns
    df = run_forecasts(df)
    df['Week'] =

    #endregion


from pandasgui.utility import fix_ipython, fix_pyqt

import numpy as np
import plotly.graph_objs as go
import dash_core_components as dcc
from threading import Thread
from pandasgui.widgets.plotly_viewer import PlotlyViewer

import dash
import dash_core_components as dcc 
import dash_html_components as html
from dash.dependencies import Input, Output

import plotly.graph_objects as go
from pandasgui.datasets import pokemon
import random

class GraphBuilder(PlotlyViewer):
    def __init__(self, df, unique_name=None):
        if not unique_name:
            unique_name = str(random.randint(0, 10000))
        p = Thread(target=self.run_plotly_dash, args=(df, unique_name, ))
        p.start()  # Start the execution
        url = f"http://127.0.0.1:8050/{unique_name}/"
        super().__init__(url=url)
        self.setWindowTitle("Graph Builder")

    def run_plotly_dash(self, df, unique_name):
        # Supress logging of HTTP requests to console
        import logging
        # log = logging.getLogger('werkzeug')
        # log.setLevel(logging.ERROR)

        external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

        app = dash.Dash(external_stylesheets=external_stylesheets,
                        # https://community.plot.ly/t/how-to-get-a-responsive-layout/18029
                        meta_tags=[
                            {"name": "viewport", "content": "width=device-width, initial-scale=1"}
                        ],
                        # https://community.plot.ly/t/multiple-dashboards/4656/3
                        name=unique_name, url_base_pathname=f'/{unique_name}/'
                        )

        app.layout = html.Div([
            dcc.Dropdown(
                id='dropdown',
                options=[{'label': col, 'value': col} for col in df.columns],
                value=df.columns[0]
            ),
            dcc.Graph(id='graph'),

        ])

        # the "value" property of the Slider is the input of the app and the output of the app is the "figure" property of the Graph.
        @app.callback(
            Output('graph', 'figure'),
            [Input('dropdown', 'value')])
        def update_figure(dropdown_value):
            return {
                'data': [go.Histogram(x=df[dropdown_value])],
                'title': unique_name
            }

        # debug=True will not work unless it's running in the main thread
        app.run_server(debug=False)


if __name__ == "__main__":
    fix_ipython()
    fix_pyqt()
    from PyQt5.QtWidgets import QApplication
    import sys

    # Create a QApplication instance or use the existing one if it exists
    app = QApplication(sys.argv)
    gb = GraphBuilder(pokemon, 'pokemon')



"""Defines sample datasets for use in testing and demos"""
import pandas as pd

__all__ = ['all_datasets', 'iris', 'mpg', 'pokemon', 'tips', 'titanic', 'flights', 'simple', 'multi_df']

iris = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/iris.csv')
mpg = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/mpg.csv')
pokemon = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/pokemon.csv')
tips = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/tips.csv')
titanic = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/titanic.csv')
flights = pd.read_csv('https://raw.githubusercontent.com/adamerose/datasets/master/flights.csv')

multi_index = pd.MultiIndex.from_tuples([('A', 'one', 'x'), ('A', 'one', 'y'), ('A', 'two', 'x'), ('A', 'two', 'y'),
                                         ('B', 'one', 'x'), ('B', 'one', 'y'), ('B', 'two', 'x'), ('B', 'two', 'y')],
                                        names=['first', 'second', 'third'])
multi_df = pd.DataFrame(pd.np.random.randn(8, 8), index=multi_index, columns=multi_index)

simple = pd.DataFrame({'a': [1, 2, 3], 'b': [10, 20, 30], 'c': [300, 200, 100]})
all_datasets = {
    "iris": iris,
    "mpg": mpg,
    "pokemon": pokemon,
    "tips": tips,
    "titanic": titanic,
    "flights": flights,
    "simple": simple,
    "multi_df": multi_df,
}



from dataclasses import dataclass


@dataclass
class Settings:
    editable: bool = True  # Are table cells editable
    block: bool = False

@dataclass
class Store:
    settings: Settings = Settings()

store = Store()


"""Defines the main PandasGUI class and related functions"""

import inspect
import sys
import os
import pkg_resources
import pandas as pd
from pydantic import BaseModel
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt
from pandasgui.widgets import PivotDialog, ScatterDialog
from pandasgui.widgets import DataFrameExplorer
from pandasgui.widgets import FindToolbar
from pandasgui.utility import fix_ipython
from pandasgui.store import store

# This makes it so PyQt5 windows don't become unresponsive in IPython outside app._exec() loops
fix_ipython()

# Provides proper stacktrace if PyQt crashes
sys.excepthook = lambda cls, exception, traceback: sys.__excepthook__(cls, exception, traceback)

# Holds references to all created PandasGUI windows so they don't get garbage collected
instance_list = []


class PandasGUI(QtWidgets.QMainWindow):

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs (): Dict of (key, value) pairs of
                         {'DataFrame name': DataFrame object}


        self.df_dicts is a dictionary of all dataframes in the GUI.
        {dataframe name: objects}

        The objects are their own dictionary of:
        {'dataframe': DataFrame object
        'dataframe_explorer': DataFrameExplorer object}

        """

        # Property initialization
        self.df_dicts = {}
        # Set in setupUI()
        self.stacked_widget = None
        self.splitter = None
        self.nav_tree = None

        # Get an application instance
        self.app = QtWidgets.QApplication.instance()
        if self.app:
            print('Using existing QApplication instance')
        if not self.app:
            self.app = QtWidgets.QApplication(sys.argv)

        super().__init__()

        # This ensures there is always a reference to this widget and it doesn't get garbage collected
        global instance_list
        instance_list.append(self)

        # https://stackoverflow.com/a/27178019/3620725
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        # Adds DataFrames listed in kwargs to df_dict.
        for i, (df_name, df_object) in enumerate(kwargs.items()):
            self.df_dicts[df_name] = {}
            self.df_dicts[df_name]['dataframe'] = df_object

        # Generates all UI contents
        self.setupUI()

        # %% Window settings
        # Set size
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        percentage_of_screen = 0.7
        size = tuple((pd.np.array([screen.width(), screen.height()]) * percentage_of_screen).astype(int))
        self.resize(QtCore.QSize(*size))
        # Center window on screen
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move(int((screen.width() - size.width()) / 2), int((screen.height() - size.height()) / 2))
        # Title and logo
        self.setWindowTitle('PandasGUI')
        pdgui_icon = 'images/icon.png'
        pdgui_icon_path = pkg_resources.resource_filename(__name__, pdgui_icon)
        self.app.setWindowIcon(QtGui.QIcon(pdgui_icon_path))

        self.show()

    def setupUI(self):
        """
        Creates and adds all widgets to GUI.
        """

        # This holds the DataFrameExplorer for each DataFrame
        self.stacked_widget = QtWidgets.QStackedWidget()

        # Make the navigation bar
        self.nav_tree = self.NavWidget(self)
        # Creates the headers.
        self.nav_tree.setHeaderLabels(['Name', 'Shape'])
        self.nav_tree.itemSelectionChanged.connect(self.nav_clicked)

        for df_name in self.df_dicts.keys():
            df_object = self.df_dicts[df_name]['dataframe']
            self.add_dataframe(df_name, df_object)

        # Make splitter to hold nav and DataFrameExplorers
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.splitter.addWidget(self.nav_tree)
        self.splitter.addWidget(self.stacked_widget)

        self.splitter.setCollapsible(0, False)
        self.splitter.setCollapsible(1, False)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)

        nav_width = self.nav_tree.sizeHint().width()
        self.splitter.setSizes([nav_width, self.width() - nav_width])
        self.splitter.setContentsMargins(10, 10, 10, 10)

        # makes the find toolbar
        self.findBar = FindToolbar(self)
        self.addToolBar(self.findBar)

        # QMainWindow setup
        self.make_menu_bar()
        self.setCentralWidget(self.splitter)

    def import_dataframe(self, path):

        if os.path.isfile(path) and path.endswith('.csv'):
            df_name = os.path.split(path)[1]
            df_object = pd.read_csv(path)
            self.add_dataframe(df_name, df_object)

        else:
            print("Invalid file: ", path)

    def add_dataframe(self, df_name, df_object):
        '''
        Add a new DataFrame to the GUI
        '''

        if type(df_object) != pd.DataFrame:
            try:
                df_object = pd.DataFrame(df_object)
                print(f'Automatically converted "{df_name}" from type {type(df_object)} to DataFrame')
            except:
                print(f'Could not convert "{df_name}" from type {type(df_object)} to DataFrame')
                return

        # Non-string column indices causes problems when pulling them from a GUI dropdown (which will give str)
        if type(df_object.columns) != pd.MultiIndex:
            df_object.columns = df_object.columns.astype(str)

        self.df_dicts[df_name] = {}
        self.df_dicts[df_name] = {}
        self.df_dicts[df_name]['dataframe'] = df_object

        dfe = DataFrameExplorer(df_object)
        self.stacked_widget.addWidget(dfe)

        self.df_dicts[df_name]['dataframe_explorer'] = dfe
        self.add_df_to_nav(df_name)

    ####################
    # Menu bar functions

    def make_menu_bar(self):
        '''
        Make the menubar and add it to the QMainWindow
        '''
        # Create a menu for setting the GUI style.
        # Uses radio-style buttons in a QActionGroup.
        menubar = self.menuBar()

        # Creates an edit menu
        editMenu = menubar.addMenu('&Edit')
        findAction = QtWidgets.QAction('&Find', self)
        findAction.setShortcut('Ctrl+F')
        findAction.triggered.connect(self.findBar.show_find_bar)
        editMenu.addAction(findAction)

        styleMenu = menubar.addMenu('&Set Style')
        styleGroup = QtWidgets.QActionGroup(styleMenu)

        # Add an option to the menu for each GUI style that exist for the user's system
        for style in QtWidgets.QStyleFactory.keys():
            styleAction = QtWidgets.QAction(f'&{style}', self, checkable=True)
            styleAction.triggered.connect(
                lambda state, style=style: self.app.setStyle(style) and self.app.setStyleSheet(""))
            styleGroup.addAction(styleAction)
            styleMenu.addAction(styleAction)
        # Set the default style
        styleAction.trigger()

        # Creates a debug menu.
        debugMenu = menubar.addMenu('&Debug')
        testDialogAction = QtWidgets.QAction('&Test', self)
        testDialogAction.triggered.connect(self.test)
        debugMenu.addAction(testDialogAction)

        '''
        # Creates a chart menu.
        chartMenu = menubar.addMenu('&Plot Charts')
        scatterDialogAction = QtWidgets.QAction('&Scatter Dialog', self)
        scatterDialogAction.triggered.connect(self.scatter_dialog)
        chartMenu.addAction(scatterDialogAction)

        # Creates a reshaping menu.
        chartMenu = menubar.addMenu('&Reshape Data')
        pivotDialogAction = QtWidgets.QAction('&Pivot Dialog', self)
        pivotDialogAction.triggered.connect(self.pivot_dialog)
        chartMenu.addAction(pivotDialogAction)
        '''

    # I just use this function for printing various things to console while the GUI is running
    def test(self):
        print('----------------')
        print('splitter', self.splitter.size())
        print('nav_tree', self.nav_tree.size())
        print('stacked_widget', self.stacked_widget.size())
        print('splitter', self.splitter.sizeHint())
        print('nav_tree', self.nav_tree.sizeHint())
        print('stacked_widget', self.stacked_widget.sizeHint())
        print('----------------')

    class NavWidget(QtWidgets.QTreeWidget):
        def __init__(self, gui):
            super().__init__()
            self.gui = gui
            self.setHeaderLabels(['HeaderLabel'])
            self.expandAll()
            self.setAcceptDrops(True)

            for i in range(self.columnCount()):
                self.resizeColumnToContents(i)

            self.setColumnWidth(0, 150)
            self.setColumnWidth(1, 150)

        def rowsInserted(self, parent: QtCore.QModelIndex, start: int, end: int):
            super().rowsInserted(parent, start, end)
            self.expandAll()

        def sizeHint(self):
            # Width
            width = 0
            for i in range(self.columnCount()):
                width += self.columnWidth(i)
            return QtCore.QSize(300, 500)

        def dragEnterEvent(self, e):
            if e.mimeData().hasUrls:
                e.accept()
            else:
                e.ignore()

        def dragMoveEvent(self, e):
            if e.mimeData().hasUrls:
                e.accept()
            else:
                e.ignore()

        def dropEvent(self, e):
            if e.mimeData().hasUrls:
                e.setDropAction(QtCore.Qt.CopyAction)
                e.accept()
                fpath_list = []
                for url in e.mimeData().urls():
                    fpath_list.append(str(url.toLocalFile()))

                for fpath in fpath_list:
                    self.gui.import_dataframe(fpath)
            else:
                e.ignore()

    def add_df_to_nav(self, df_name, parent=None):
        """
        Add DataFrame to the nav by looking up the DataFrame by name in df_dicts

        Args:
            df_name (str): Name of the DataFrame
            parent (QTreeWidgetItem): Parent item in the nav tree hierarchy
        """

        if parent is None:
            parent = self.nav_tree

        # Calculate and format the shape of the DataFrame
        shape = self.df_dicts[df_name]['dataframe'].shape
        shape = str(shape[0]) + ' X ' + str(shape[1])

        item = QtWidgets.QTreeWidgetItem(parent, [df_name, shape])
        self.nav_tree.itemSelectionChanged.emit()
        self.nav_tree.setCurrentItem(item)

    def nav_clicked(self):
        """
        Show the DataFrameExplorer corresponding to the highlighted nav item.
        """
        try:
            item = self.nav_tree.selectedItems()[0]
        except IndexError:
            return

        df_name = item.data(0, Qt.DisplayRole)

        dfe = self.df_dicts[df_name]['dataframe_explorer']
        self.stacked_widget.setCurrentWidget(dfe)

    ####################
    # Dialog functions. TODO: Rewrite these all

    def pivot_dialog(self):
        default = self.nav_tree.currentItem().data(0, Qt.DisplayRole)
        win = PivotDialog(self.df_dicts, default=default, gui=self)

    def scatter_dialog(self):
        default = self.nav_tree.currentItem().data(0, Qt.DisplayRole)
        win = ScatterDialog(self.df_dicts, default=default, gui=self)


def show(*args, settings: dict = {}, **kwargs):
    """
    Create and show a PandasGUI window with all the DataFrames passed. *args and **kwargs should all be DataFrames

    Args:
        *args: These should all be DataFrames. The GUI uses stack inspection to get the variable name to use in the GUI
        block (bool): Indicates whether to run app._exec on the PyQt application to block further execution of script
        **kwargs: These should all be DataFrames. The key is the desired name and the value is the DataFrame object
    """

    for key, value in settings.items():
        setattr(store.settings, key, value)

    # Get the variable names in the scope show() was called from
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()

    # Make a dictionary of the DataFrames from the position args and get their variable names using inspect
    dataframes = {}
    for i, df_object in enumerate(args):
        df_name = 'untitled' + str(i + 1)

        for var_name, var_val in callers_local_vars:
            if var_val is df_object:
                df_name = var_name

        dataframes[df_name] = df_object

    # Add the dictionary of positional args to the kwargs
    if (any([key in kwargs.keys() for key in dataframes.keys()])):
        print("Warning! Duplicate DataFrame names were given, duplicates were ignored.")
    kwargs = {**kwargs, **dataframes}

    pandas_gui = PandasGUI(**kwargs)

    if store.settings.block:
        pandas_gui.app.exec_()


if __name__ == '__main__':

    # Fix lack of stack trace on PyQt exceptions
    def my_exception_hook(exctype, value, traceback):
        # Print the error and traceback
        print(exctype, value, traceback)
        # Call the normal Exception hook after
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)


    sys.excepthook = my_exception_hook

    try:
        # Get paths of drag & dropped files and prepare to open them in the GUI
        file_paths = sys.argv[1:]
        if file_paths:
            file_dataframes = {}
            for path in file_paths:
                if os.path.isfile(path) and path.endswith('.csv'):
                    df = pd.read_csv(path)
                    filename = os.path.split(path)[1]
                    file_dataframes[filename] = df
            show(**file_dataframes)

        # Script was run normally, open sample data sets
        else:
            from pandasgui.datasets import all_datasets
            show(**all_datasets, settings={'block': True})

    # Catch errors and call input() so they can be viewed before the console window closes when running with drag n drop
    except Exception as e:
        print(e)
        import traceback

        traceback.print_exc()



import os, sys
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QUrl
import tempfile
from bokeh.resources import CDN
# TODO: use local static JS instead of loading from CDN
from bokeh.embed import file_html


instance_list = []

# If QtWebEngineWidgets is imported while a QApplication instance already exists it will fail, so we have to
try:
    from PyQt5 import QtWebEngineWidgets
except ImportError as e:
    if e.msg == "QtWebEngineWidgets must be imported before a QCoreApplication instance is created":
        print("Killing QApplication to reimport QtWebEngineWidgets")
        from PyQt5 import sip

        app = QApplication.instance()
        app.quit()
        sip.delete(app)
        del app
        from PyQt5 import QtWebEngineWidgets

        # Without remaking the QApplication you will get an unrecoverable crash on the next attempted usage of it
        app = QApplication([])
    else:
        raise e


class BokehViewer(QtWebEngineWidgets.QWebEngineView):
    def __init__(self, fig):
        # Create a QApplication instance or use the existing one if it exists
        self.app = QApplication.instance() if QApplication.instance() else QApplication(sys.argv)

        super().__init__()

        # This ensures there is always a reference to this widget and it doesn't get garbage collected
        global instance_list
        instance_list.append(self)

        # Store the Plotly html in a temp file to load into the webview.
        # NOTE - QWebEngineView.setHtml doesn't work with data over 2 MB, that's why I use a temp file instead
        self.temp_file = tempfile.TemporaryFile(mode='w', suffix='.html')

        self.update(fig)

        self.setWindowTitle("Bokeh Viewer")
        self.resize(640, 480)
        self.show()

    def update(self, fig):
        html = file_html(fig, resources=CDN, title="my fig")
        self.temp_file.seek(0)
        self.temp_file.truncate()
        self.temp_file.write(html)
        self.temp_file.seek(0)
        self.load(QUrl.fromLocalFile(self.temp_file.name))


if __name__ == "__main__":
    from bokeh.plotting import figure
    from pandasgui.utility import fix_ipython
    fix_ipython()

    fig = figure(sizing_mode = 'stretch_both')
    fig.circle([1, 2, 3, 4], [3, 4, 5, 6], size=20)

    bv = BokehViewer(fig)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(name='pandasgui',
      version='0.1.15',
      description='A GUI for Pandas DataFrames.',
      author='Adam Rose',
      author_email='adam.e.rose@hotmail.com',
      url='https://github.com/adamerose/pandasgui',
      packages=find_packages(),
      include_package_data=True,
      long_description=long_description,
      long_description_content_type='text/markdown',
      install_requires=[
          'pandas',
          'PyQt5',
          'PyQtWebEngine',
          'seaborn',
          'plotly',
          'setuptools']
      )



from pandasgui.utility import fix_ipython, fix_pyqt

from pandasgui.widgets.bokeh_viewer import BokehViewer
from pandasgui.widgets.plotly_viewer import PlotlyViewer
from pandasgui.utility import flatten_multiindex
from pandasgui.datasets import pokemon
from PyQt5 import QtWidgets
import plotly.express as px

# from plotly.offline import plot
# fig = px.histogram(pokemon, x="HP")
# plot(px.histogram(pokemon, x="HP"))

class GraphBuilder(QtWidgets.QWidget):
    def __init__(self, df):
        df.columns = flatten_multiindex(df.columns)

        super().__init__()
        self.setWindowTitle("Graph Builder")
        self.df = df.copy()

        self.picker = QtWidgets.QComboBox()
        self.picker.addItems(df.columns)
        self.picker.currentIndexChanged.connect(self.update_plot)
        self.figure_viewer = PlotlyViewer()

        self.layout = QtWidgets.QVBoxLayout()

        self.layout.addWidget(self.picker)
        self.layout.addWidget(self.figure_viewer)

        self.setLayout(self.layout)
        self.update_plot()

    def update_plot(self):
        col = self.picker.currentText()
        fig = px.histogram(self.df, x=col)
        # if self.df[col].dtype.name in ['object', 'bool', 'category']:
        self.figure_viewer.set_figure(fig)

if __name__ == "__main__":
    fix_ipython()
    fix_pyqt()
    from PyQt5.QtWidgets import QApplication
    import sys

    # Create a QApplication instance or use the existing one if it exists
    app = QApplication(sys.argv)
    gb = GraphBuilder(pokemon)
